# -*- coding: utf-8 -*-
"""
Created on Fri Jan 22 15:48:34 2020

@author: Suhas Somnath, Sudhajit Misra, Rama Vasudevan
"""

from __future__ import division, print_function, absolute_import, unicode_literals
import sys
import numpy as np  # For array operations
import sidpy as sid
from sidpy.sid import Reader
import re
from os import path

if sys.version_info.major == 3:
    unicode = str

class AscReader(Reader):
    """
    Reads Scanning Tunnelling Spectroscopy (STS) data in .asc files obtained from Omicron STMs.
    These are actually compatible with Nanonis controllers for Omicron microscopes,
    It seems they are probably NOT compatible with matrix controller outputs.
    See the group discussion at https://groups.google.com/g/pycroscopy/c/d7lcGnlPFpo/m/UEysliC8CAAJ

    TODO: At this point this AscReader does not appear to work for non-Z spec files. Needs to be fixed!
    """

    def read(self, verbose=False, parm_encoding='utf-8'):
        """
        Reads the file given in file_path into a sidpy dataset

        Parameters
        ----------
        verbose : Boolean (Optional)
            Whether or not to show  print statements for debugging
        parm_encoding : str, optional
            Codec to be used to decode the bytestrings into Python strings if
            needed. Default 'utf-8'

        Returns
        -------
        sidpy.Dataset : List of sidpy.Dataset objects.
            Multi-channel inputs are separated into individual dataset objects
        """

        file_path = self._input_file_path
        folder_path, file_name = path.split(file_path)
        file_name = file_name[:-4]

        # Extracting the raw data into memory
        file_handle = open(file_path, 'r')
        string_lines = file_handle.readlines()
        file_handle.close()

        # Extract parameters from the first few header lines
        parm_dict, num_headers = self._read_parms(string_lines)
        if verbose: print(parm_dict)
        num_rows = int(parm_dict['Main-y_pixels'])
        num_cols = int(parm_dict['Main-x_pixels'])
        num_pos = num_rows * num_cols
        spectra_length = int(parm_dict['Main-z_points'])

        # Extract the STS data from subsequent lines
        raw_data_2d = self._read_data(string_lines, num_pos, spectra_length, num_headers)
        raw_data = np.reshape(raw_data_2d, (num_rows, num_cols, spectra_length))

        # Generate the x / voltage / spectroscopic axis:
        volt_vec = np.linspace(parm_dict['Spectroscopy-Device_1_Start [Volt]'],
                               parm_dict['Spectroscopy-Device_1_End [Volt]'],
                               spectra_length)

        # Build row, column vectors
        xvec = np.linspace(0, parm_dict['Main-x_length'], num_cols)
        yvec = np.linspace(0, parm_dict['Main-y_length'], num_rows)

        # now write it to the sidpy dataset object
        data_set = sid.Dataset.from_array(raw_data, name='Omicron_STS')
        data_set.data_type = 'spectral_image'

        # Add quantity and units
        data_set.units = parm_dict['Main-value_unit']
        data_set.quantity = 'Current'

        # Add dimension info
        data_set.set_dimension(0, sid.Dimension(xvec,
                                                name='x',
                                                units='nm',
                                                quantity='position',
                                                dimension_type=sid.DimensionType.SPATIAL))

        data_set.set_dimension(1, sid.Dimension(yvec,
                                                name='y',
                                                units='nm',
                                                quantity='position',
                                                dimension_type=sid.DimensionType.SPATIAL))

        data_set.set_dimension(2, sid.Dimension(volt_vec, name='I',
                                                units=parm_dict['Main-value_unit'], quantity='Current',
                                                dimension_type=sid.DimensionType.SPECTRAL))

        # append metadata
        data_set.metadata = parm_dict

        # Return the sidy dataset
        return data_set

    def _read_data(self, string_lines, num_pos, spectra_length, num_headers):
        """
        Reads the data from lines of the data file

        Parameters
        ----------
        string_lines : :class:`list` of str
            Lines containing the data in string format, separated by tabs
        num_pos : unsigned int
            Number of pixels
        spectra_length : unsigned int
            Number of points in the spectral / voltage axis
        num_headers : unsigned int
            Number of header lines to ignore

        Returns
        -------
        raw_data_2d : 2D numpy array
            Data arranged as [position x voltage points]
        """
        raw_data_2d = np.zeros(shape=(num_pos, spectra_length), dtype=np.float32)
        for line_ind in range(num_pos):
            this_line = string_lines[num_headers + line_ind]
            string_spectrum = this_line.split('\t')[:-1]  # omitting the new line
            raw_data_2d[line_ind] = np.array(string_spectrum, dtype=np.float32)
        return raw_data_2d

    def _parse_file_path(self, input_path):
        pass

    @staticmethod
    def _read_parms(string_lines):
        """
        Returns the parameters regarding the experiment as dictionary

        Parameters
        ----------
        string_lines : list of strings
            Lines from the data file in string representation

        Returns
        -------
        parm_dict : dictionary
            Dictionary of parameters regarding the experiment
        """
        # Reading parameters stored in the first few rows of the file

        def parse_header(line):
            # ".  Aux2_V:"
            match_obj = re.match(r'\.  (.*):', line, re.M | re.I)
            type_list = [str]
            if match_obj:
                raw_vals = [type_caster(match_obj.group(ind)) for ind, type_caster in
                            zip(range(1, 1 + len(type_list)), type_list)]
                return raw_vals[0]
            else:
                return None

        def parse_parm(line):
            # ".  .  Auto_Flush_Period = 0.1 Second"
            match_obj = re.match(r'\.  \.  (.*) = (.*)', line, re.M | re.I)
            type_list = [str, str]
            if match_obj:
                raw_vals = [type_caster(match_obj.group(ind)) for ind, type_caster in
                            zip(range(1, 1 + len(type_list)), type_list)]

                # Some cleaning:
                raw_vals[0] = raw_vals[0].replace('-', '_').strip()  # We use '-' as a level separator
                raw_vals[1] = raw_vals[1].replace('--', '').strip()

                # often, units are on the values side, see if these can be transitioned over to the key:
                vals_split = raw_vals[1].split(' ')
                if len(vals_split) == 2:
                    raw_vals = [raw_vals[0] + ' [' + vals_split[1] + ']', vals_split[0]]
                try:
                    raw_vals[1] = float(raw_vals[1])
                    # convert those values that should be integers:
                    if raw_vals[1] % 1 == 0:
                        raw_vals[1] = int(raw_vals[1])
                except ValueError:
                    pass
                return {raw_vals[0]: raw_vals[1]}
            else:
                return None

        def flatten_dict(nested_dict, separator='-'):
            # From: https://codereview.stackexchange.com/questions/21033/flatten-dictionary-in-python-functional-style
            def expand(outer_key, outer_value):
                if isinstance(outer_value, dict):
                    return [(outer_key + separator + inner_key, inner_value) for inner_key, inner_value in flatten_dict(outer_value).items()]
                else:
                    return [(outer_key, outer_value)]

            items = [item for outer_key, outer_value in nested_dict.items() for item in expand(outer_key, outer_value)]

            return dict(items)

        # #############################################################################################################

        temp_dict = dict()

        line = string_lines[1]
        if line.startswith('# Created by SPIP'):
            line = line.replace('# Created by SPIP ', '').replace('\n', ' ')
            ind = line.index(' ')
            temp_dict['SPIP_version'] = line[:ind]
            temp_dict['creation_time'] = line[ind + 1:]

        # #################################################################################################

        line_offset = 3
        for line_ind, line in enumerate(string_lines[line_offset:]):
            if parse_header(line) is not None:
                line_offset += line_ind
                break
            line = line.replace('# ', '')
            line = line.replace('\n', '')
            temp = line.split('=')
            try:
                test = temp[1].strip()
                try:
                    test = float(test)
                    # convert those values that should be integers:
                    if test % 1 == 0:
                        test = int(test)
                except ValueError:
                    pass
            except IndexError:
                pass

            temp_dict[temp[0].strip().replace('-', '_')] = test

        main_dict = {'Main': temp_dict.copy()}

        # #################################################################################################

        curr_cat_name = None
        temp_dict = dict()

        for ind, line in enumerate(string_lines[line_offset:]):
            if line.strip().startswith('# Start of Data:'):
                line_offset += ind + 1
                break
            header_name = parse_header(line)
            if header_name:
                if curr_cat_name is not None:
                    main_dict[curr_cat_name] = temp_dict.copy()
                    temp_dict = dict()
                curr_cat_name = header_name
            else:
                this_parm = parse_parm(line)
                if this_parm is None:
                    continue
                temp_dict.update(this_parm)
        if len(temp_dict) > 0:
            main_dict[curr_cat_name] = temp_dict.copy()

        # #################################################################################################

        return flatten_dict(main_dict), line_offset

    def can_read(self):
        """
        Tests whether or not the provided file has a .asc extension
        Returns
        -------

        """

        return super(AscReader, self).can_read(extension='asc')



from .igor_ibw import IgorIBWReader
from .AR_hdf5  import ARhdf5Reader
from .bruker_nano import BrukerAFMReader

__all__ = ['IgorIBWReader', 'ARhdf5Reader', 'BrukerAFMReader']
all_readers = [IgorIBWReader, ARhdf5Reader, BrukerAFMReader]


# -*- coding: utf-8 -*-
"""
SpeReader is designed to read a single Raman spectra from a Princeton Instruments CCD camera.
The reader takes in the Princeton Instruments .SPE file and produces a sidpy dataset.

Created on Fri May 14 2020

@author: Sumner Harris
"""

import numpy as np
import re
from sidpy import Dataset, Dimension, Reader

class RamanSpeReader(Reader):
    
    def read(self):
        """
        Reads a single Raman spectra from a Princeton Instruments CCD camera.

        Returns
        -------
        data_set: sidpy.Dataset object
            wraps all the raw data and metadata from the input file into a Dataset object
        
        """
        # open the .SPE file
        with open(self._input_file_path, 'rb') as f:
            lines = f.readlines()
            # Create an empty dictionary for the metadata
            metadata_dictionary = {}

            # Search through the file for the needed metadata
            metadata_dictionary['date_acquired'] = re.search(b'date="(.*?)"', lines[1])[1].decode('ANSI')  
            metadata_dictionary['width'] = int(re.search(b'width="(.*?)"', lines[1])[1])
            metadata_dictionary['height'] = int(re.search(b'height="(.*?)"', lines[1])[1])
            metadata_dictionary['size'] = metadata_dictionary['width']*metadata_dictionary['height']
            metadata_dictionary['exposure_time'] = int(re.search(b'<ExposureTime type="Double">(.*?)</ExposureTime>', lines[1])[1])
            metadata_dictionary['excitation_wavelength'] = float(re.search(b'laserLine="(.*?)"',lines[1])[1])
            metadata_dictionary['center_wavelength'] = float(re.search(b'<CenterWavelength type="Double">(.*?)</CenterWavelength>',lines[1])[1])
            metadata_dictionary['orientation'] = re.search(b'orientation="(.*?)"',lines[1])[1].decode('ANSI')

            # Get the wavelength and intensity
            wavelength_string = re.search(b'<Wavelength xml:space="preserve">(.*?)</Wavelength>',lines[1])[1].decode('utf-8')
            wavelength = np.array(wavelength_string.split(','), dtype=np.float64)

            f.seek(4100)
            intensity = np.fromfile(f,dtype=np.float32,count=metadata_dictionary['size'])

            raman_shift_wavenumbers = 1e7*(1/metadata_dictionary['excitation_wavelength'] - 1/wavelength)

            f.close()
            
        # create the sidpy dataset
        data_set = Dataset.from_array(intensity, name='Raman Spectra')

        data_set.data_type = 'spectrum'
        data_set.units = 'counts'
        data_set.quantity = 'Intensity'

        # set dimensions
        data_set.set_dimension(0, Dimension(raman_shift_wavenumbers, name='Raman Shift',
                                                 units = 'cm-1',
                                                 quantity='Raman shift',
                                                 dimension_type='spectral'))
        data_set.set_dimension(1, Dimension(intensity, name='Intensity',
                                                 units = 'counts',
                                                 quantity='intensity',
                                                 dimension_type='spectral'))        

        data_set.metadata = metadata_dictionary

        return data_set
    
    def can_read(self):
        """
        Tests whether or not the provided file has .spe extension

        Returns
        -------
        """

        return super(RamanSpeReader, self).can_read(extension='spe')



from .omicron_asc import AscReader
from .nanonis_dat import NanonisDatReader
from .nanonis_sxm import NanonisSXMReader
from .nanonis_3ds import Nanonis3dsReader

__all__ = ['AscReader', 'NanonisDatReader', 'Nanonis3dsReader', 'NanonisSXMReader']
all_readers = [AscReader, NanonisDatReader, Nanonis3dsReader, NanonisSXMReader]


"""
:class:`~ScopeReaders.generic.image.ImageTranslator` class that extracts the
data from typical image files into :class:`~sidpy.Dataset` objects

Created on Feb 9, 2016

@author: Suhas Somnath, Chris Smith
"""

from __future__ import division, print_function, absolute_import, unicode_literals

import os
import sys
import numpy as np
from PIL import Image
from sidpy.base.num_utils import contains_integers
from sidpy import Dataset, Dimension, Reader

if sys.version_info.major == 3:
    unicode = str
else:
    FileExistsError = ValueError
    FileNotFoundError = ValueError


class ImageReader(Reader):
    """
    Translates data from an image file to an HDF5 file
    """

    def __init__(self, *args, **kwargs):
        super(ImageReader, self).__init__(*args, **kwargs)

    @staticmethod
    def _parse_file_path(image_path):
        """
        Returns a list of all files in the directory given by path

        Parameters
        ---------------
        image_path : str
            absolute path to the image file

        Returns
        ----------
        image_path : str
            Absolute file path to the image
        h5_path : str
            absolute path to the desired output HDF5 file.
        """
        if not isinstance(image_path, (str, unicode)):
            raise TypeError("'image_path' argument for ImageTranslator should be a str or unicode")
        if not os.path.exists(os.path.abspath(image_path)):
            raise FileNotFoundError('Specified image does not exist.')
        else:
            image_path = os.path.abspath(image_path)

        return image_path

    def read(self, bin_factor=None, interp_func=Image.BICUBIC,
             normalize=False, **image_args):
        """
        Translates the image in the provided file into a USID HDF5 file

        Parameters
        ----------------
        bin_factor : uint or array-like of uint, optional
            Down-sampling factor for each dimension.  Default is None.
            If specifying different binning for each dimension, please specify as (height binning, width binning)
        interp_func : int, optional. Default = :attr:`PIL.Image.BICUBIC`
            How the image will be interpolated to provide the down-sampled or binned image.
            For more information see instructions for the `resample` argument for :meth:`PIL.Image.resize`
        normalize : boolean, optional. Default = False
            Should the raw image be normalized between the values of 0 and 1
        image_args : dict
            Arguments to be passed to read_image.  Arguments depend on the type of image.

        Returns
        ----------

        """
        image_path = self._parse_file_path(self._input_file_path)

        image = read_image(image_path, **image_args)
        image_parms = dict()
        usize, vsize = image.shape[:2]

        '''
        Check if a bin_factor is given.  Set up binning objects if it is.
        '''
        if bin_factor is not None:
            if isinstance(bin_factor, (list, tuple)):
                if not contains_integers(bin_factor, min_val=1):
                    raise TypeError('bin_factor should contain positive whole integers')
                if len(bin_factor) == 2:
                    bin_factor = tuple(bin_factor)
                else:
                    raise ValueError('Input parameter `bin_factor` must be a length 2 array-like or an integer.\n' +
                                     '{} was given.'.format(bin_factor))

            elif isinstance(bin_factor, int):
                bin_factor = (bin_factor, bin_factor)
            else:
                raise TypeError('bin_factor should either be an integer or an iterable of positive integers')

            if np.min(bin_factor) < 0:
                raise ValueError('bin_factor must consist of positive factors')

            if interp_func not in [Image.NEAREST, Image.BILINEAR, Image.BICUBIC, Image.LANCZOS]:
                raise ValueError("'interp_func' argument for ImageTranslator.translate must be one of "
                                 "PIL.Image.NEAREST, PIL.Image.BILINEAR, PIL.Image.BICUBIC, PIL.Image.LANCZOS")

            image_parms.update({'image_binning_size': bin_factor, 'image_PIL_resample_mode': interp_func})
            usize = int(usize / bin_factor[0])
            vsize = int(vsize / bin_factor[1])

            # Unfortunately, we need to make a round-trip through PIL for the interpolation. Not possible with numpy
            img_obj = Image.fromarray(image)
            img_obj = img_obj.resize((vsize, usize), resample=interp_func)
            image = np.asarray(img_obj)

        # Working around occasional "cannot modify read-only array" error
        image = image.copy()

        '''
        Normalize Raw Image
        '''
        if normalize:
            image -= np.min(image)
            image = image / np.float32(np.max(image))

        image_parms.update({'normalized': normalize,
                            'image_min': np.min(image), 'image_max': np.max(image)})

        data_set = Dataset.from_array(image, name='random')

        data_set.data_type = 'image'
        data_set.units = 'a. u.'
        data_set.quantity = 'Intensity'

        data_set.set_dimension(0,
                               Dimension(np.arange(usize), 'y', units='a. u.',
                                         quantity='Length',
                                         dimension_type='spatial'))
        data_set.set_dimension(1,
                               Dimension(np.arange(vsize), 'x', units='a. u.',
                                         quantity='Length',
                                         dimension_type='spatial'))

        return data_set

    def can_read(self):
        """
        Tests whether or not the provided file has a .ndata extension
        Returns
        -------

        """
        exts = ['jpg', 'jpeg', 'png', 'tiff', 'bmp', 'csv', 'txt']
        return super(ImageReader, self).can_read(extension=exts)


def read_image(image_path, as_grayscale=True, as_numpy_array=True, *args, **kwargs):
    """
    Read the image file at `image_path` into a numpy array either via numpy (.txt) or via pillow (.jpg, .tif, etc.)

    Parameters
    ----------
    image_path : str
        Path to the image file
    as_grayscale : bool, optional. Default = True
        Whether or not to read the image as a grayscale image
    as_numpy_array : bool, optional. Default = True
        If set to True, the image is read into a numpy array. If not, it is returned as a pillow Image

    Returns
    -------
    image : :class:`numpy.ndarray` or :class:`PIL.Image.Image`
        if `as_numpy_array` is set to True - Array containing the image from the file `image_path`.
        If `as_numpy_array` is set to False - PIL.Image object containing the image within the file - `image_path`.
    """
    ext = os.path.splitext(image_path)[-1]
    if ext in ['.txt', '.csv']:
        if ext == '.csv' and 'delimiter' not in kwargs.keys():
            kwargs['delimiter'] = ','
        img_data = np.loadtxt(image_path, *args, **kwargs)
        if as_numpy_array:
            return img_data
        else:
            img_obj = Image.fromarray(img_data)
            img_obj = img_obj.convert(mode="L")
            return img_obj
    else:
        img_obj = Image.open(image_path)
        if as_grayscale:
            img_obj = img_obj.convert(mode="L", **kwargs)

        if as_numpy_array:
            # Open the image as a numpy array
            return np.asarray(img_obj)
        return img_obj



# -*- coding: utf-8 -*-
"""
Created on Fri Dec 10 15:48:00 2021

@author: Rama Vasudevan, based on a translator in legacy pycroscopy
which was created on Fri May 25 16:04:34 2016 by  Suhas Somnath

"""

from os import stat
import sys
import numpy as np  # For array operations
import sidpy as sid
from sidpy.sid import Reader
from .base_utils import read_binary_data

from collections import OrderedDict

class BrukerAFMReader(Reader):
    """
    Extracts data and metadata from Bruker AFM (.nnn) files containing
    images or force curves where n is an integer

    """

    def read(self, verbose=False, parm_encoding='utf-8'):
        """
        Reads the file given in file_path into a sidpy dataset

        Parameters
        ----------
        verbose : Boolean (Optional)
            Whether or not to show  print statements for debugging
        parm_encoding : str, optional
            Codec to be used to decode the bytestrings into Python strings if
            needed. Default 'utf-8'

        Returns
        -------
        sidpy.Dataset : List of sidpy.Dataset objects.
            Multi-channel inputs are separated into individual sidpy dataset objects
        """
        self.file_path = self._input_file_path
        self.meta_data, other_parms = self._extract_metadata()

        type_suffixes = ['Image', 'Force_Curve', 'Force_Map']
        # 0 - stack of scan images
        # 1 - single force curve
        # 2 - force map
        force_count = 0
        image_count = 0
        for class_name in self.meta_data.keys():
            if 'Ciao force image list' in class_name:
                force_count += 1
            elif 'Ciao image list' in class_name:
                image_count += 1
        data_type = 0
        if force_count > 0:
            if image_count > 0:
                data_type = 2
            else:
                data_type = 1

        global_parms = dict()
        global_parms['data_type'] = 'Bruker_AFM_' + type_suffixes[data_type]
        global_parms['translator'] = 'Bruker_AFM'

        flat_dict = dict()
        for class_name, sub_dict in other_parms.items():
            for key, val in sub_dict.items():
                flat_dict[class_name + '_' + key] = val

        trans_funcs = [self._read_image_stack, self._read_force_curve, self._read_force_map]
        dataset = trans_funcs[data_type]()

        return dataset

    def _read_force_curve(self):
        """
        Reads the force curves from the proprietary file and writes them to sidpy dataset object
        
        """
       
        # Find out the size of the force curves from the metadata:
        layer_info = None
        for class_name in self.meta_data.keys():
            if 'Ciao force image list' in class_name:
                layer_info = self.meta_data[class_name]
                break
        
        tr_rt = [int(item) for item in layer_info['Samps/line'].split(' ')]
        m=0
        datasets = []
        titles = []
        metadata = []
        for class_name in self.meta_data.keys():
            if 'Ciao force image list' in class_name:
                layer_info = self.meta_data[class_name]
                quantity = layer_info.pop('Image Data_4')
                title = quantity.split("\"")[1]
                data = self._read_data_vector(layer_info)
                self.data = data
                data_split =  np.split(data, len(data)//tr_rt[m])
                titles.append((title, quantity))
                datasets.append(data_split)
                metadata.append(layer_info)
                m+=1

        xdata = datasets[1] #not sure if this is right but let's go with it.
        title = titles[0][0]
        quantity = titles[0][1]
        
        sid_datasets = []

        for k in range(len(datasets[0])):
            zdata = datasets[0][k]
            xdata = datasets[1][k]
            
            data_set = sid.Dataset.from_array(zdata, title=title)
            data_set.data_type = 'Spectrum'

            #Add quantity and units
            data_set.units = 'nm' #check this one
            data_set.quantity = quantity

            #Add dimension info
            data_set.set_dimension(0, sid.Dimension(xdata,
                                                    name = 'z',
                                                    units='nm', quantity = 'z',
                                                    dimension_type='spectral'))
            
            data_set.original_metadata = metadata[k]
            sid_datasets.append(data_set)

        return sid_datasets

    def _read_image_stack(self):
        """
        Reads the scan images from the proprietary file and puts them into sidpy.Dataset
        Parameters
        ----------

        """
      
        # Find out the size of the force curves from the metadata:
        layer_info = None
        for class_name in self.meta_data.keys():
            if 'Ciao image list' in class_name:
                layer_info = self.meta_data[class_name]
                break
        
        #Here are teh dimension details
        num_samps_line = layer_info['Samps/line']
        num_lines = layer_info['Number of lines']
       
        #Read through and write to sidpy dataset objects
        datasets = []
        for class_name in self.meta_data.keys():
            if 'Ciao image list' in class_name:
                layer_info = self.meta_data[class_name]
                #print(layer_info)
                quantity = layer_info.pop('Image Data_2')
                title = quantity.split("\"")[1]
                data = self._read_image_layer(layer_info)
                num_cols, num_rows = data.shape
                data_set = sid.Dataset.from_array(data, title=title)
                data_set.data_type = 'Image'

                #Add quantity and units
                data_set.units = 'nm' #check this one
                data_set.quantity = quantity

                #Add dimension info
                data_set.set_dimension(0, sid.Dimension(np.linspace(0, num_samps_line, num_cols),
                                                        name = 'x',
                                                        units='nm', quantity = 'x',
                                                        dimension_type='spatial'))
                data_set.set_dimension(1, sid.Dimension(np.linspace(0, num_lines, num_rows),
                                                        name = 'y',
                                                        units='nm', quantity='y',
                                                        dimension_type='spatial'))

                # append metadata
                data_set.original_metadata = self.meta_data[class_name]
                data_set.data_type = 'image'

                datasets.append(data_set)

        return datasets
        

    def _read_force_map(self):
        """
        Reads the scan image + force map from the proprietary file and writes it to sidpy datasets
        Parameters
        ----------
        h5_meas_grp : h5py.Group object
            Reference to the measurement group
        """
        # First lets write the image into the measurement group that has already been created:
        image_parms = self.meta_data['Ciao image list']
        quantity = image_parms.pop('Image Data_2')
        image_mat = self._read_image_layer(image_parms)
        #h5_chan_grp = create_indexed_group(h5_meas_grp, 'Channel')
        #write_main_dataset(h5_chan_grp, np.reshape(image_mat, (-1, 1)), 'Raw_Data',
        #                   # Quantity and Units needs to be fixed by someone who understands these files better
        #                   quantity, 'a. u.',
        #                   [Dimension('X', 'nm', image_parms['Samps/line']),
        #                    Dimension('Y', 'nm', image_parms['Number of lines'])],
        #                   Dimension('single', 'a. u.', 1), dtype=np.float32, compression='gzip')
        # Think about standardizing attributes for rows and columns
        #write_simple_attrs(h5_chan_grp, image_parms)

        # Now work on the force map:
        force_map_parms = self.meta_data['Ciao force image list']
        quantity = force_map_parms.pop('Image Data_4')
        force_map_vec = self._read_data_vector(force_map_parms)
        tr_rt = [int(item) for item in force_map_parms['Samps/line'].split(' ')]
        force_map_2d = force_map_vec.reshape(image_mat.size, np.sum(tr_rt))
        #h5_chan_grp = create_indexed_group(h5_meas_grp, 'Channel')
        #write_main_dataset(h5_chan_grp, force_map_2d, 'Raw_Data',
        #                   # Quantity and Units needs to be fixed by someone who understands these files better
        #                   quantity, 'a. u.',
        #                   [Dimension('X', 'nm', image_parms['Samps/line']),
        #                    Dimension('Y', 'nm', image_parms['Number of lines'])],
        #                   Dimension('Z', 'nm', int(np.sum(tr_rt))), dtype=np.float32, compression='gzip')
        # Think about standardizing attributes
        #write_simple_attrs(h5_chan_grp, force_map_parms)

        return force_map_2d

    def _extract_metadata(self):
        """
        Reads the metadata in the header
        Returns
        -------
        meas_parms : OrderedDict
            Ordered dictionary of Ordered dictionaries (one per image / force channel, etc.)
        other_parms : OrderedDict
            Ordered Dictionary of Ordered dictionaries containing all other metadata
        """
        other_parms = OrderedDict()
        meas_parms = OrderedDict()
        curr_category = ''
        temp_dict = OrderedDict()
        with open(self.file_path, "rb") as file_handle:
            for ind, line in enumerate(file_handle):
                line = line.decode("utf-8", 'ignore')
                trimmed = line.strip().replace("\\", "").replace('@', '')
                split_data = trimmed.split(':')

                # First account for wierdly formatted metadata that
                if len(split_data) == 3:
                    split_data = [split_data[1] + '_' + split_data[0], split_data[-1]]
                elif len(split_data) > 3:
                    # Date:
                    split_ind = trimmed.index(':')
                    split_data = [trimmed[:split_ind], trimmed[split_ind + 1:]]

                # At this point, split_data should only contain either 1 (class header) or 2 elements
                if len(split_data) == 1:
                    if len(temp_dict) > 0:
                        if 'Ciao image list' in curr_category or 'Ciao force image list' in curr_category:
                            # In certain cases the same class name occurs multiple times.
                            # Append suffix to existing name and to this name
                            count = 0
                            for class_name in meas_parms.keys():
                                if curr_category in class_name:
                                    count += 1
                            if count == 0:
                                meas_parms[curr_category] = temp_dict.copy()
                            else:
                                if count == 1:
                                    for class_name in meas_parms.keys():
                                        if curr_category == class_name:
                                            # Remove and add back again with suffix
                                            # This should only ever happen once.
                                            # The next time we come across the same class, all elements already have
                                            # suffixes
                                            meas_parms[curr_category + '_0'] = meas_parms.pop(curr_category)
                                            break
                                meas_parms[curr_category + '_' + str(count)] = temp_dict.copy()
                        else:
                            curr_category = curr_category.replace('Ciao ', '')
                            other_parms[curr_category] = temp_dict.copy()

                    if "*File list end" in trimmed:
                        break
                    curr_category = split_data[0].replace('*', '')
                    temp_dict = OrderedDict()
                elif len(split_data) == 2:
                    split_data = [item.strip() for item in split_data]
                    try:
                        num_val = float(split_data[1])
                        if int(num_val) == num_val:
                            num_val = int(num_val)
                        temp_dict[split_data[0]] = num_val
                    except ValueError:
                        temp_dict[split_data[0]] = split_data[1]
                else:
                    print(split_data)

        return meas_parms, other_parms

    def _read_data_vector(self, layer_info):
        """
        Reads data relevant to a single image, force curve, or force map
        Parameters
        ----------
        layer_info : OrderedDictionary
            Parameters describing the data offset, length and precision in the binary file
        Returns
        -------
        data_vec : np.ndarray
            1D array containing data represented by binary data
        """
        data_vec = read_binary_data(self.file_path, layer_info['Data offset'], layer_info['Data length'],
                                    layer_info['Bytes/pixel'])

        # Remove translation specific values from dictionary:
        for key in ['Data offset', 'Data length', 'Bytes/pixel']:
            _ = layer_info.pop(key)
        return data_vec
    
    def _read_image_layer(self, layer_info):
        """
        Reads a single scan image layer / channel
        Parameters
        ----------
        layer_info : OrderedDictionary
            Parameters describing the data offset, length and precision in the binary file
        Returns
        -------
        data_mat : numpy.ndarray
            2D array representing the requested channel of information
        """
        data_vec = self._read_data_vector(layer_info)
        data_mat = data_vec.reshape(layer_info['Number of lines'], layer_info['Samps/line'])
        return data_mat
    
    def can_read(self):
        """
        Tests whether or not the provided file has a suitable extension
        Returns
        -------
        not used for now...
        
        """
        return super(BrukerAFMReader, self).can_read(extension='')




# -*- coding: utf-8 -*-
"""
Created on Fri Mar 12 15:39:00 202`0`

@author: Rama Vasudevan
"""

import sys
import numpy as np  # For array operations
import sidpy as sid
from sidpy.sid import Reader
import re
from os import path


class NanonisDatReader(Reader):
    """
    Reads files obtained via Nanonis controllers in .dat files.
    These are generally point spectroscopy measurements.

    """

    def read(self, verbose=False):
        """
        Reads the file given in file_path into a sidpy dataset

        Parameters
        ----------
        verbose : Boolean (Optional)
            Whether or not to show  print statements for debugging

        Returns
        -------
        sidpy.Dataset : List of sidpy.Dataset objects.
            Multi-channel inputs are separated into individual dataset objects
        """

        file_path = self._input_file_path
        folder_path, file_name = path.split(file_path)

        # Extracting the raw data into memory
        file_handle = open(file_path, 'r')
        string_lines = file_handle.readlines()
        file_handle.close()

        data_start = string_lines.index('[DATA]\n')
        header = string_lines[:data_start - 1]

        channel_names_with_units = string_lines[data_start+1].split('\t')
        channel_names = [channel_names_with_units[ind].split('(')[0] for ind in range (len(channel_names_with_units))]
        channel_units = [re.search(r'\((.*?)\)',channel_names_with_units[ind]).group(1) for ind in range (len(channel_names_with_units))]

        # Extract parameters from the first few header lines
        parm_dict = self._read_parms(header)

        if verbose:
            print('Found parameters dictionary {}'.format(parm_dict))

        # Extract the STS data from subsequent lines
        raw_data = np.loadtxt(file_path, skiprows=data_start+2)
        if verbose:
            print('Read data of shape {}'.format(raw_data.shape))

        # Generate the x / voltage ß/ spectroscopic axis:
        volt_vec = raw_data[:,0]
        if verbose:
            print('Found spectroscopic vector of size {}'.format(volt_vec.shape))
            print('Spectroscopy vector has title {}'.format(channel_names[0]))
            print('Spectrsocopy vector values: {}'.format(volt_vec))
        datasets = [] #list of sidpy datasets that will be output

        # Add quantity and units
        for chan_ind, chan_name in enumerate(channel_names[1:]): #start from 1 because 0th column is the spectral one

            if verbose:
                print('Making sidpy dataset with channel {}'.format(chan_name))

            # now write it to the sidpy dataset object
            data_set = sid.Dataset.from_array(raw_data[:,chan_ind+1], name=chan_name)
            data_set.data_type = 'spectrum'
            data_set.units = channel_units[chan_ind+1]
            data_set.quantity = chan_name

            # Add dimension info
            data_set.set_dimension(0, sid.Dimension(volt_vec, name=chan_name,
                                                    units=channel_units[0], quantity='Voltage',
                                                    dimension_type=sid.DimensionType.SPECTRAL))

            # append metadata
            data_set.metadata = parm_dict
            datasets.append(data_set)

        # Return the sidy dataset
        return datasets

    @staticmethod
    def _read_parms(header):
        """
        Returns the parameters regarding the experiment as dictionary

        Parameters
        ----------
        string_lines : list of strings
            Lines from the data file in string representation

        Returns
        -------
        parm_dict : dictionary
            Dictionary of parameters regarding the experiment
        """
        # Reading parameters stored in the first few rows of the file
        # "Look through the header and create a dictionary from it"
        parm_dict = {}
        for line in header:
            vals = line.split('\t')
            key = vals[0]
            if len(vals[1:-1]) == 1:
                try:
                    #If the key, value pair is a float, convert it
                    val = float(vals[1:-1][0])
                except ValueError:
                    val = vals[1:-1][0]
            elif len(vals[1:-1]) == 0:
                val = []
            else:
                val = vals[1:-1][0]
            parm_dict[key] = val
        return parm_dict

    def can_read(self):
        """
        Tests whether or not the provided file has a .asc extension
        Returns
        -------

        """

        return super(NanonisDatReader, self).can_read(extension='dat')



#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-

################################################################################
# Python class for reading FEI Velox .emd files into sidpy Dataset
# and extracting all metadata
#
# Written by Gerd Duscher, UTK 2021
#
# Works for python 3
#
################################################################################
import json
import h5py
import sys
import numpy as np

import sidpy

__all__ = ["EMDReader", "version"]

version = '0.1beta'


debugLevel = 0  # 0=none, 1-3=basic, 4-5=simple, 6-10 verbose

if sys.version_info.major == 3:
    unicode = str


class EMDReader(sidpy.Reader):
    def __init__(self, file_path):
        """
        Creates an instance of EMDReader which can read one or more HDF5
        datasets formatted in the FEI Velox style EDM file

        We can read Images, and SpectrumStreams (SpectrumImages and Spectra).
        Please note that all original metadata are retained in each sidpy dataset.

        Parameters
        ----------
        file_path : str
            Path to a HDF5 file
        """

        super(EMDReader, self).__init__(file_path)

        # Let h5py raise an OS error if a non-HDF5 file was provided
        self._h5_file = h5py.File(file_path, mode='r+')

        self.datasets = []
        self.data_array = None
        self.metadata = None
        self.number_of_frames = 1

    def can_read(self):
        """
        Checks whether or not this Reader can read the provided file

        Returns
        -------
        bool :
            True if this Reader can read the provided file
            Else, False
        """
        if 'Application' in self._h5_file:
            return 'Velox' in self._h5_file['Application']
        else:
            return False

    def read(self, eds_stream=False):
        """
        Reads all available datasets in FEI Velox style hdf5 files with .edm

        Parameters
        ----------
        eds_stream: boolean
            switch to return spectrum image (default - False) or original spectrum stream (True)

        Returns
        -------
        datasets: list of sidpy.Dataset objects
            Datasets present in the provided file
        """

        if 'Data' not in self._h5_file:
            raise TypeError('Velox EMD File is empty')
        for key in self._h5_file['Data']:
            if key == 'Image':
                for image_key in self._h5_file['Data']['Image']:
                    self.get_data('Data/Image/' + image_key)
                    self.get_image()
                    self.get_metadata(-1)
            elif key == 'SpectrumStream':
                for stream_key in self._h5_file['Data']['SpectrumStream']:
                    self.get_data('Data/SpectrumStream/' + stream_key)
                    self.get_eds(eds_stream)
                    self.get_metadata(-1)

        self.close()
        return self.datasets

    def get_data(self, image_key):
        self.data_array = self._h5_file[image_key]['Data']
        metadata_array = self._h5_file[image_key]['Metadata'][:, 0]
        metadata_string = metadata_array.tobytes().decode("utf-8")
        self.metadata = dict(json.loads(metadata_string.rstrip('\x00')))
        if 'AcquisitionSettings' in self._h5_file[image_key]:
            self.metadata['AcquisitionSettings'] = json.loads(self._h5_file[image_key]['AcquisitionSettings'][0])

    def get_eds(self, eds_stream=False):
        if 'AcquisitionSettings' not in self.metadata:
            eds_stream = True
        if eds_stream:
            self.datasets.append(sidpy.Dataset.from_array(self.data_array))
        else:
            data_array = self.get_eds_spectrum()
            if data_array.shape[0] == 1 and data_array.shape[1] == 1:
                data_array = np.array(data_array).flatten()
            self.datasets.append(sidpy.Dataset.from_array(data_array))
        # print(self.datasets[-1])

        self.datasets[-1].original_metadata = self.metadata

        detectors = self.datasets[-1].original_metadata['Detectors']
        if eds_stream:
            pass
        else:
            offset = 0.
            dispersion = 1.
            for detector in detectors.values():
                if self.metadata['BinaryResult']['Detector'] in detector['DetectorName']:
                    if 'OffsetEnergy' in detector:
                        offset = float(detector['OffsetEnergy'])
                    if 'Dispersion' in detector:
                        dispersion = float(detector['Dispersion'])

            self.datasets[-1].units = 'counts'
            self.datasets[-1].quantity = 'intensity'
            energy_scale = np.arange(self.datasets[-1].shape[-1]) * dispersion + offset

            if self.datasets[-1].ndim == 1:
                self.datasets[-1].data_type = 'spectrum'

                self.datasets[-1].set_dimension(0, sidpy.Dimension(energy_scale,
                                                                   name='energy_scale', units='eV',
                                                                   quantity='energy',
                                                                   dimension_type='spectral'))

            else:
                self.datasets[-1].data_type = 'spectral_image'
                self.datasets[-1].set_dimension(2, sidpy.Dimension(energy_scale,
                                                                   name='energy_scale', units='eV',
                                                                   quantity='energy',
                                                                   dimension_type='spectral'))
                scale_x = float(self.metadata['BinaryResult']['PixelSize']['width']) * 1e9
                scale_y = float(self.metadata['BinaryResult']['PixelSize']['height']) * 1e9

                self.datasets[-1].set_dimension(0, sidpy.Dimension(np.arange(self.datasets[-1].shape[0]) * scale_x,
                                                                   name='x', units='nm',
                                                                   quantity='distance',
                                                                   dimension_type='spatial'))
                self.datasets[-1].set_dimension(1, sidpy.Dimension(np.arange(self.datasets[-1].shape[1]) * scale_y,
                                                                   name='y', units='nm',
                                                                   quantity='distance',
                                                                   dimension_type='spatial'))

    def get_eds_spectrum(self):
        acquisition = self.metadata['AcquisitionSettings']
        print(acquisition)
        size_x = 1
        size_y = 1
        if 'RasterScanDefinition' in acquisition:
            size_x = int(acquisition['RasterScanDefinition']['Width'])
            size_y = int(acquisition['RasterScanDefinition']['Height'])
        spectrum_size = int(acquisition['bincount'])

        self.number_of_frames = int(np.ceil((self.data_array[:, 0] == 65535).sum() / (size_x * size_y)))
        # print(size_x,size_y,number_of_frames)
        data = np.zeros((size_x * size_y, spectrum_size))
        # progress = tqdm(total=number_of_frames)
        pixel_number = 0
        frame = 0
        for value in self.data_array[:, 0]:
            if value == 65535:
                pixel_number += 1
                if pixel_number >= size_x * size_y:
                    pixel_number = 0
                    frame += 1
                    # print(frame)
                    # progress.update(1)
            else:
                data[pixel_number, value] += 1
        self.number_of_frames = frame
        return np.reshape(data, (size_x, size_y, spectrum_size))

    def get_image(self):

        scale_x = float(self.metadata['BinaryResult']['PixelSize']['width']) * 1e9
        scale_y = float(self.metadata['BinaryResult']['PixelSize']['height']) * 1e9

        if self.data_array.shape[2] == 1:
            self.datasets.append(sidpy.Dataset.from_array(self.data_array[:, :, 0]))
            self.datasets[-1].data_type = 'image'
            self.datasets[-1].set_dimension(0, sidpy.Dimension(np.arange(self.data_array.shape[0]) * scale_x,
                                                               name='x', units='nm',
                                                               quantity='distance',
                                                               dimension_type='spatial'))
            self.datasets[-1].set_dimension(1, sidpy.Dimension(np.arange(self.data_array.shape[1]) * scale_y,
                                                               name='y', units='nm',
                                                               quantity='distance',
                                                               dimension_type='spatial'))
        else:
            self.data_array = np.moveaxis(self.data_array, source=[0, 1, 2], destination=[1, 2, 0])
            self.datasets.append(sidpy.Dataset.from_array(self.data_array))
            self.datasets[-1].data_type = 'image_stack'
            self.datasets[-1].set_dimension(0, sidpy.Dimension(np.arange(self.data_array.shape[0]),
                                                               name='frame', units='frame',
                                                               quantity='time',
                                                               dimension_type='temporal'))
            self.datasets[-1].set_dimension(1, sidpy.Dimension(np.arange(self.data_array.shape[1]) * scale_x,
                                                               name='x', units='nm',
                                                               quantity='distance',
                                                               dimension_type='spatial'))
            self.datasets[-1].set_dimension(2, sidpy.Dimension(np.arange(self.data_array.shape[2]) * scale_y,
                                                               name='y', units='nm',
                                                               quantity='distance',
                                                               dimension_type='spatial'))
        self.datasets[-1].original_metadata = self.metadata

        self.datasets[-1].units = 'counts'
        self.datasets[-1].quantity = 'intensity'

    def get_metadata(self, index):
        metadata = self.datasets[index].original_metadata
        experiment = {'detector': metadata['BinaryResult']['Detector'],
                      'acceleration_voltage': float(metadata['Optics']['AccelerationVoltage']),
                      'microscope': metadata['Instrument']['InstrumentClass'],
                      'start_date_time': int(metadata['Acquisition']['AcquisitionStartDatetime']['DateTime'])}

        if metadata['Optics']['ProbeMode'] == "1":
            experiment['probe_mode'] = "convergent"
            if 'BeamConvergence' in metadata['Optics']:
                experiment['convergence_angle'] = float(metadata['Optics']['BeamConvergence'])
        else:  # metadata['Optics']['ProbeMode'] == "2":
            experiment['probe_mode'] = "parallel"
            experiment['convergence_angle'] = 0.0
        experiment['stage'] = {"holder": "",
                               "position": {"x": float(metadata['Stage']['Position']['x']),
                                            "y": float(metadata['Stage']['Position']['y']),
                                            "z": float(metadata['Stage']['Position']['z'])},
                               "tilt": {"alpha": float(metadata['Stage']['AlphaTilt']),
                                        "beta": float(metadata['Stage']['BetaTilt'])}}

        self.datasets[index].metadata['experiment'] = experiment

    def close(self):
        self._h5_file.close()



# -*- coding: utf-8 -*-
"""
Created on Mon Sep 13 21:46:14 2021

@author: Raj
"""

import sidpy as sid
from sidpy.sid import Reader
from sidpy.sid import Dimension

import os
import numpy as np
import h5py

from pyNSID.io.hdf_io import write_nsid_dataset
from pyNSID.io.hdf_io import create_indexed_group, write_simple_attrs

class PiFMTranslator(Reader):
    """
    Class that writes images, spectrograms, point spectra and associated ancillary data sets to h5 file in pyUSID data
    structure.
    """

    def read(self ):
        """
        Parameters
        ----------
        file_path : String / unicode
            Absolute path of the .ibw file
        verbose : Boolean (Optional)
            Whether or not to show  print statements for debugging

        Returns
        -------
        sidpy.Dataset : List of sidpy.Dataset objects.
            Image layers are saved as separate Dataset objects
        """
        
        self.get_path()
        self.read_anfatec_params()
        self.read_file_desc()
        self.read_spectrograms()
        self.read_imgs()
        self.read_spectra()
        self.datasets = self.make_datasets()

        return self.datasets

    def create_h5(self, append_path='', overwrite=False):
        """
        Writes a new HDF5 file with the translated data
        
        append_path : string (Optional)
            h5_file to add these data to, must be a path to the h5_file on disk
        overwrite : bool (optional, default=False)
            If True, will overwrite an existing .h5 file of the same name        
        """
        self.create_hdf5_file(append_path, overwrite)
        self.write_datasets_hdf5()
        
        return
        
    def get_path(self):
        """writes full path, directory, and file name as attributes to class"""
        
        self.path = self._input_file_path
        full_path = os.path.realpath(self.path)
        directory = os.path.dirname(full_path)
        # file name
        basename = os.path.basename(self.path)
        self.full_path = full_path
        self.directory = directory
        self.basename = basename

    def read_anfatec_params(self):
        """reads the scan parameters and writes them to a dictionary"""
        
        params_dictionary = {}
        params = True
        with open(self.path, 'r', encoding="ISO-8859-1") as f:
            for line in f:
                if params:
                    sline = [val.strip() for val in line.split(':')]
                    if len(sline) == 2 and sline[0][0] != ';':
                        params_dictionary[sline[0]] = sline[1]
                    #in ANFATEC parameter files, all attributes are written before file references.
                    if sline[0].startswith('FileDesc'):
                        params = False
            f.close()
        self.params_dictionary = params_dictionary
        self.x_len, self.y_len = int(params_dictionary['xPixel']), int(params_dictionary['yPixel'])

    def read_file_desc(self):
        """reads spectrogram, image, and spectra file descriptions and stores all to dictionary where
        the key:value pairs are filename:[all descriptors]"""
        
        spectrogram_desc = {}
        img_desc = {}
        spectrum_desc = {}
        pspectrum_desc = {}
        
        with open(self.path,'r', encoding="ISO-8859-1") as f:

            lines = f.readlines()
            for index, line in enumerate(lines):

                sline = [val.strip() for val in line.split(':')]
                #if true, then file describes image.

                if sline[0].startswith('FileDescBegin'):
                    no_descriptors = 5
                    file_desc = []

                    for i in range(no_descriptors):

                        line_desc = [val.strip() for val in lines[index+i+1].split(':')]
                        file_desc.append(line_desc[1])
                    #img_desc['filename'] = caption, scale, physical unit, offset
                    img_desc[file_desc[0]] = file_desc[1:]

                #if true, file describes spectrogram (ie hyperspectral image)
                if sline[0].startswith('FileDesc2Begin'):

                    no_descriptors = 10
                    file_desc = []

                    for i  in range(no_descriptors):
                        line_desc = [val.strip() for val in lines[index+i+1].split(':')]
                        file_desc.append(line_desc[1])
                    #caption, bytes perpixel, scale, physical unit, offset, offset, datatype, bytes per reading
                    #filename wavelengths, phys units wavelengths.
                    spectrogram_desc[file_desc[0]] = file_desc[1:]

                if sline[0].startswith('AFMSpectrumDescBegin'):

                    file_desc = []
                    line_desc = [val.strip() for val in lines[index+1].split(':')][1]

                    if 'powerspectrum' in line_desc:
                        no_descriptors = 2

                        for i in range(no_descriptors):

                            line_desc = [val.strip() for val in lines[index+i+1].split(':')]
                            file_desc.append(line_desc[1])
                        #file name, position x, position y
                        pspectrum_desc[file_desc[0]] = file_desc[1:]

                    else:

                        no_descriptors = 7
                        for i in range(no_descriptors):
                            line_desc = [val.strip() for val in lines[index+i+1].split(':')]
                            file_desc.append(line_desc[1])
                        #file name, position x, position y
                        spectrum_desc[file_desc[0]] = file_desc[1:]
            f.close()
            
        self.img_desc = img_desc
        self.spectrogram_desc = spectrogram_desc
        self.spectrum_desc = spectrum_desc
        self.pspectrum_desc = pspectrum_desc
        
    def read_spectrograms(self):
        """reads spectrograms, associated spectral values, and saves them in two dictionaries"""
        
        spectrograms = {}
        spectrogram_spec_vals = {}
        
        for file_name, descriptors in self.spectrogram_desc.items():

            spec_vals_i = np.loadtxt(os.path.join(self.directory, file_name.strip('.int') + 'Wavelengths.txt'))
            #if true, data is acquired with polarizer, with an attenuation data column
         
            if np.array(spec_vals_i).ndim == 2:
                spectrogram_spec_vals[file_name] = spec_vals_i[:, 0]
                attenuation = {}
                attenuation[file_name] = spec_vals_i[:, 1]
                self.attenuation = attenuation
            
            else:
                spectrogram_spec_vals[file_name] = spec_vals_i
                
            #load and save spectrograms
            spectrogram_i = np.fromfile(os.path.join(self.directory, file_name), dtype='i4')
            spectrograms[file_name] = np.zeros((self.x_len, self.y_len, len(spec_vals_i)))
            
            for y, line in enumerate(np.split(spectrogram_i, self.y_len)):

                for x, pt_spectrum in enumerate(np.split(line, self.x_len)):
                    spectrograms[file_name][x, y, :] = pt_spectrum * float(descriptors[2])

        self.spectrograms = spectrograms
        self.spectrogram_spec_vals = spectrogram_spec_vals

    def read_imgs(self):
        """reads images and saves to dictionary"""
        imgs = {}
        
        for file_name, descriptors in self.img_desc.items():

            img_i = np.fromfile(os.path.join(self.directory, file_name), dtype='i4')
            imgs[file_name] = np.zeros((self.x_len, self.y_len))
        
            for y, line in enumerate(np.split(img_i, self.y_len)):
            
                for x, pixel in enumerate(np.split(line, self.x_len)):
                    imgs[file_name][x, y] = pixel * float(descriptors[1])
        
        self.imgs = imgs

    def read_spectra(self):
        """reads all point spectra and saves to dictionary"""
        spectra = {}
        spectra_spec_vals = {}
        spectra_x_y_dim_name = {}
        
        for file_name, descriptors in self.spectrum_desc.items():
            spectrum_f = np.loadtxt(os.path.join(self.directory, file_name), skiprows=1)
            spectra_spec_vals[file_name] = spectrum_f[:, 0]
            spectra[file_name] = spectrum_f[:,1]
        
            with open(os.path.join(self.directory, file_name)) as f:
                spectra_x_y_dim_name[file_name]  = f.readline().strip('\n').split('\t')
        
        for file_name, descriptors in self.pspectrum_desc.items():
            spectrum_f = np.loadtxt(os.path.join(self.directory, file_name), skiprows=1)
            spectra_spec_vals[file_name] = spectrum_f[:, 0]
            spectra[file_name] = spectrum_f[:,1]
        
            with open(os.path.join(self.directory, file_name)) as f:
                spectra_x_y_dim_name[file_name]  = f.readline().strip('\n').split('\t')
        
        self.spectra = spectra
        self.spectra_spec_vals = spectra_spec_vals
        self.spectra_x_y_dim_name = spectra_x_y_dim_name

    def make_datasets(self):
        
        datasets = []
        self.make_dimensions()
 
        # Spectrograms
        if bool(self.spectrogram_desc):
           
           for spectrogram_f, descriptors in self.spectrogram_desc.items():

               # channel_i = create_indexed_group(self.h5_meas_grp, 'Channel_')
               spec_vals_i = self.spectrogram_spec_vals[spectrogram_f]
               
               spectrogram_data = self.spectrograms[spectrogram_f]
               dset = sid.Dataset.from_array(spectrogram_data, name=descriptors[0])
               dset.data_type = 'Spectrogram'
               dset.set_dimension(0, self.dim0)
               dset.set_dimension(1, self.dim0)

               # spectrogram_spec_dims = Dimension('Wavelength', descriptors[8], spec_vals_i)
               spectrogram_dims = Dimension(values=spec_vals_i, name='Spectrogram', 
                                            units=descriptors[3], quantity='Wavelength', type='spectral' )
               dset.set_dimension(2, spectrogram_dims)
               dset.metadata = {'Caption': descriptors[0],
                                'Bytes_Per_Pixel': descriptors[1],
                                'Scale': descriptors[2],
                                'Physical_Units': descriptors[3],
                                'Offset': descriptors[4],
                                'Datatype': descriptors[5],
                                'Bytes_Per_Reading': descriptors[6],
                                'Wavelength_File': descriptors[7],
                                'Wavelength_Units': descriptors[8]}
               
               datasets.append(dset)
               
        # Images
        if bool(self.img_desc):
            
            for img_f, descriptors in self.img_desc.items():
                
                img_data = self.imgs[img_f]
                dset = sid.Dataset.from_array(img_data, name = descriptors[0])
                dset.data_type = 'Image'
                dset.set_dimension(0, self.dim0)
                dset.set_dimension(1, self.dim1)   
                dset.units = descriptors[2]
                dset.quantity = descriptors[0]
                dset.metadata =  {'Caption': descriptors[0],
                                  'Scale': descriptors[1],
                                  'Physical_Units': descriptors[2],
                                  'Offset': descriptors[3]}

                datasets.append(dset)
                
        # Spectra
        if bool(self.spectrum_desc):
            
            for spec_f, descriptors in self.spectrum_desc.items():
                
                #create new measurement group for each spectrum
                x_name = self.spectra_x_y_dim_name[spec_f][0].split(' ')[0]
                x_unit = self.spectra_x_y_dim_name[spec_f][0].split(' ')[1]
                y_name = self.spectra_x_y_dim_name[spec_f][1].split(' ')[0]
                y_unit = self.spectra_x_y_dim_name[spec_f][1].split(' ')[1]
                dset = sid.Dataset.from_array(self.spectra[spec_f], name = 'Raw_Spectrum')

                dset.set_dimension(0, Dimension(np.array([float(descriptors[1])]), 
                                                name='X',units=self.params_dictionary['XPhysUnit'].replace('\xb5','u'),
                                                quantity = 'X_position'))
                
                dset.set_dimension(1, Dimension(np.array([float(descriptors[2])]), 
                                                name='Y',units=self.params_dictionary['YPhysUnit'].replace('\xb5','u'),
                                                quantity = 'Y_position'))
                dset.data_type = 'Spectrum'
                dset.units = y_unit
                dset.quantity = y_name
                spectra_dims = Dimension(values=self.spectra_spec_vals[spec_f], name='Wavelength', 
                                        units=x_unit, quantity=x_name, type='spectral' )
                dset.set_dimension(2, spectra_dims)
                dset.metadata = {'XLoc': descriptors[1], 'YLoc': descriptors[2]}

                datasets.append(dset)
        
        # Power Spectra
        if bool(self.pspectrum_desc):
            for spec_f, descriptors in self.pspectrum_desc.items():

                #create new measurement group for each spectrum
                x_name = self.spectra_x_y_dim_name[spec_f][0].split(' ')[0]
                x_unit = self.spectra_x_y_dim_name[spec_f][0].split(' ')[1]
                y_name = self.spectra_x_y_dim_name[spec_f][1].split(' ')[0]
                y_unit = self.spectra_x_y_dim_name[spec_f][1].split(' ')[1]
                dset = sid.Dataset.from_array(self.spectra[spec_f], name = 'Power_Spectrum')

                dset.set_dimension(0, Dimension(np.array([0]), 
                                                name='X',units=self.params_dictionary['XPhysUnit'].replace('\xb5','u'),
                                                quantity = 'X_position'))
                
                dset.set_dimension(1, Dimension(np.array([0]), 
                                                name='Y',units=self.params_dictionary['YPhysUnit'].replace('\xb5','u'),
                                                quantity = 'Y_position'))
                dset.data_type = 'Spectrum'
                dset.units = y_unit
                dset.quantity = y_name
                spectra_dims = Dimension(values=self.spectra_spec_vals[spec_f], name='Wavelength', 
                                        units=x_unit, quantity=x_name, type='spectral' )
                dset.set_dimension(2, spectra_dims)
                dset.metadata = {'XLoc': 0, 'YLoc': 0}
                datasets.append(dset)                

        return datasets

    def make_dimensions(self):
        x_range = float(self.params_dictionary['XScanRange'])
        y_range = float(self.params_dictionary['YScanRange'])
        x_center = float(self.params_dictionary['xCenter'])
        y_center = float(self.params_dictionary['yCenter'])

        x_start = x_center-(x_range/2); x_end = x_center+(x_range/2)
        y_start = y_center-(y_range/2); y_end = y_center+(y_range/2)

        dx = x_range/self.x_len
        dy = y_range/self.y_len
        
        #assumes y scan direction:down; scan angle: 0 deg
        y_linspace = -np.arange(y_start, y_end, step=dy)
        x_linspace = np.arange(x_start, x_end, step=dx)
        
        qtyx = self.params_dictionary['XPhysUnit'].replace('\xb5', 'u')
        qtyy = self.params_dictionary['YPhysUnit'].replace('\xb5', 'u')
        
        self.dim0 = Dimension(x_linspace, name = 'x', units = qtyx, 
                              dimension_type = 'spatial', quantity='Length')
        self.dim1 = Dimension(y_linspace, name = 'y', units = qtyy, 
                              dimension_type = 'spatial', quantity='Length')
        
        # self.pos_ind, self.pos_val, self.pos_dims = pos_ind, pos_val, pos_dims
        
        return 
        
    # HDF5 creation        
    def create_hdf5_file(self, append_path='', overwrite=False):
        """ Sets up the HDF5 file for writing 
        
        append_path : string (Optional)
            h5_file to add these data to, must be a path to the h5_file on disk
        overwrite : bool (optional, default=False)
            If True, will overwrite an existing .h5 file of the same name        
        """
        
        if not append_path:
            h5_path = os.path.join(self.directory, self.basename.replace('.txt', '.h5'))
            if os.path.exists(h5_path):
                if not overwrite:
                    raise FileExistsError('This file already exists). Set attribute overwrite to True')
                else:
                    print('Overwriting file', h5_path)
                    #os.remove(h5_path)
                    
            self.h5_f = h5py.File(h5_path, mode='w')

        else:
            if not os.path.exists(append_path):
                raise Exception('File does not exist. Check pathname.')
            self.h5_f = h5py.File(append_path, mode='r+')

        self.h5_img_grp = create_indexed_group(self.h5_f, "Images")
        self.h5_spectra_grp = create_indexed_group(self.h5_f, "Spectra")
        self.h5_spectrogram_grp = create_indexed_group(self.h5_f, "Spectrogram")
        
        write_simple_attrs(self.h5_img_grp, self.params_dictionary)
        write_simple_attrs(self.h5_spectra_grp, self.params_dictionary)
        write_simple_attrs(self.h5_spectrogram_grp, self.params_dictionary)
        
        return
    
    def write_datasets_hdf5(self):
        """ Writes the datasets as pyNSID datasets to the HDF5 file"""
        for dset in self.datasets:
            
            if 'IMAGE' in dset.data_type.name:
        
                write_nsid_dataset(dset, self.h5_img_grp)
        
            elif 'SPECTRUM' in dset.data_type.name:
        
                write_nsid_dataset(dset, self.h5_spectra_grp) 
                
            else:
                
                write_nsid_dataset(dset, self.h5_spectrogram_grp) 
    
        self.h5_f.file.close()    
    
        return


# -*- coding: utf-8 -*-
"""
Created on Wed Dec 07 16:04:34 2016

Updated on Fri Sep 11 15:22:00 2020 for ScopeReader

@author: Suhas Somnath, Chris R. Smith, Raj Giridhargopal, Rama Vasudevan
"""

import sys
import numpy as np  # For array operations
import sidpy as sid
from sidpy.sid import Reader

try:
    from igor import binarywave as bw
except ModuleNotFoundError:
    bw = None


class IgorIBWReader(Reader):
    """
    Extracts data and metadata from Igor Binary Wave (.ibw) files containing
    images or force curves
    """

    def read(self, verbose=False, parm_encoding='utf-8'):
        """
        Reads the file given in file_path into a sidpy dataset

        Parameters
        ----------
        verbose : Boolean (Optional)
            Whether or not to show  print statements for debugging
        parm_encoding : str, optional
            Codec to be used to decode the bytestrings into Python strings if
            needed. Default 'utf-8'

        Returns
        -------
        sidpy.Dataset : List of sidpy.Dataset objects.
            Multi-channel inputs are separated into individual dataset objects
        """
        file_path = self._input_file_path

        # Load the ibw file first
        ibw_obj = bw.load(file_path)
        ibw_wave = ibw_obj.get('wave')
        parm_dict = self._read_parms(ibw_wave, parm_encoding)
        chan_labels, chan_units = self._get_chan_labels(ibw_wave, parm_encoding)

        if verbose:
            print('Channels and units found:')
            print(chan_labels)
            print(chan_units)

        # Get the data to figure out if this is an image or a force curve
        images = ibw_wave.get('wData')

        datasets = [] #list of sidpy datasets

        if images.shape[-1] != len(chan_labels):
            chan_labels = chan_labels[1:]  # for layer 0 null set errors in older AR software
            chan_units = chan_units[1:]

        if images.ndim == 3:  # Image stack
            if verbose:
                print('Found image stack of size {}'.format(images.shape))

            num_rows = parm_dict['ScanLines']
            num_cols = parm_dict['ScanPoints']

            for channel in range(images.shape[-1]):
                #Convert it to sidpy dataset object
                data_set = sid.Dataset.from_array(images[:,:,channel], title=chan_labels[channel])
                data_set.data_type = 'Image'

                #Add quantity and units
                data_set.units = chan_units[channel]
                data_set.quantity = chan_labels[channel]

                #Add dimension info
                data_set.set_dimension(0, sid.Dimension(np.linspace(0, parm_dict['FastScanSize'], num_cols),
                                                        name = 'x',
                                                        units=chan_units[channel], quantity = 'x',
                                                        dimension_type='spatial'))
                data_set.set_dimension(1, sid.Dimension(np.linspace(0, parm_dict['SlowScanSize'], num_rows),
                                                        name = 'y',
                                                        units=chan_units[channel], quantity='y',
                                                        dimension_type='spatial'))

                # append metadata
                data_set.original_metadata = parm_dict
                data_set.data_type = 'image'

                #Finally, append it
                datasets.append(data_set)

        else:  # single force curve
            if verbose:
                print('Found force curve of size {}'.format(images.shape))

            images = np.atleast_3d(images)  # now [Z, chan, 1]

            # Find the channel that corresponds to either Z sensor or Raw:
            try:
                chan_ind = chan_labels.index('ZSnsr')
                spec_data = images[:,chan_ind].squeeze()
            except ValueError:
                try:
                    chan_ind = chan_labels.index('Raw')
                    spec_data = images[:,chan_ind,0].squeeze()
                except ValueError:
                    # We don't expect to come here. If we do, spectroscopic values remains as is
                    spec_data = np.arange(images.shape[2])

            #Go through the channels
            for channel in range(images.shape[-2]):

                # The data generated above varies linearly. Override.
                # For now, we'll shove the Z sensor data into the spectroscopic values.

                #convert to sidpy dataset
                data_set = sid.Dataset.from_array((images[:,channel,0]), title=chan_labels[channel])

                if verbose:
                    print('Channel {} and spec_data is {}'.format(channel, spec_data))

                #Set units, quantity
                data_set.units = chan_units[channel]
                data_set.quantity = chan_labels[channel]

                data_set.set_dimension(0, sid.Dimension(spec_data, name = chan_labels[channel],
                                                        units=chan_units[channel], quantity=chan_labels[channel],
                                                        dimension_type='spectral'))

                #append metadata
                data_set.data_type = 'line_plot'
                data_set.original_metadata = parm_dict

                #Add dataset to list
                datasets.append(data_set)

        # Return the dataset
        return datasets

    @staticmethod
    def _read_parms(ibw_wave, codec='utf-8'):
        """
        Parses the parameters in the provided dictionary

        Parameters
        ----------
        ibw_wave : dictionary
            Wave entry in the dictionary obtained from loading the ibw file
        codec : str, optional
            Codec to be used to decode the bytestrings into Python strings if needed.
            Default 'utf-8'

        Returns
        -------
        parm_dict : dictionary
            Dictionary containing parameters
        """
        parm_string = ibw_wave.get('note')
        if type(parm_string) == bytes:
            try:
                parm_string = parm_string.decode(codec)
            except:
                parm_string = parm_string.decode('ISO-8859-1')  # for older AR software
        parm_string = parm_string.rstrip('\r')
        parm_list = parm_string.split('\r')
        parm_dict = dict()
        for pair_string in parm_list:
            temp = pair_string.split(':')
            if len(temp) == 2:
                temp = [item.strip() for item in temp]
                try:
                    num = float(temp[1])
                    parm_dict[temp[0]] = num
                    try:
                        if num == int(num):
                            parm_dict[temp[0]] = int(num)
                    except OverflowError:
                        pass
                except ValueError:
                    parm_dict[temp[0]] = temp[1]

        # Grab the creation and modification times:
        other_parms = ibw_wave.get('wave_header')
        for key in ['creationDate', 'modDate', 'bname']:
            try:
                parm_dict[key] = other_parms[key]
            except KeyError:
                pass
        return parm_dict

    @staticmethod
    def _get_chan_labels(ibw_wave, codec='utf-8'):
        """
        Retrieves the names of the data channels and default units

        Parameters
        ----------
        ibw_wave : dictionary
            Wave entry in the dictionary obtained from loading the ibw file
        codec : str, optional
            Codec to be used to decode the bytestrings into Python strings if needed.
            Default 'utf-8'

        Returns
        -------
        labels : list of strings
            List of the names of the data channels
        default_units : list of strings
            List of units for the measurement in each channel
        """
        temp = ibw_wave.get('labels')
        labels = []
        for item in temp:
            if len(item) > 0:
                labels += item
        for item in labels:
            if item == '':
                labels.remove(item)

        default_units = list()
        for chan_ind, chan in enumerate(labels):
            # clean up channel names
            if type(chan) == bytes:
                chan = chan.decode(codec)
            if chan.lower().rfind('trace') > 0:
                labels[chan_ind] = chan[:chan.lower().rfind('trace') + 5]
            else:
                labels[chan_ind] = chan
            # Figure out (default) units
            if chan.startswith('Phase'):
                default_units.append('deg')
            elif chan.startswith('Current'):
                default_units.append('A')
            else:
                default_units.append('m')

        return labels, default_units

    def can_read(self):
        """
        Tests whether or not the provided file has a .ibw extension
        Returns
        -------

        """

        return super(IgorIBWReader, self).can_read(extension='ibw')



# -*- coding: utf-8 -*-
"""
Created on Fri Nov 5 16:43:00 2021

@author: Rama Vasudevan
"""
import numpy as np  # For array operations
import sidpy as sid
from sidpy.sid import Reader, Dimension, DimensionType

try:
    import nanonispy as nap
except ModuleNotFoundError:
    nap = None


class NanonisSXMReader(Reader):

    @staticmethod
    def _parse_sxm_parms(header_dict, signal_dict):
        """
        Parse sxm files.
        Parameters
        ----------
        header_dict : dict
        signal_dict : dict
        Returns
        -------
        parm_dict : dict
        """
        parm_dict = dict()
        data_dict = dict()

        # Create dictionary with measurement parameters
        meas_parms = {key: value for key, value in header_dict.items()
                      if value is not None}
        info_dict = meas_parms.pop('data_info')
        parm_dict['meas_parms'] = meas_parms

        # Create dictionary with channel parameters
        channel_parms = dict()
        channel_names = info_dict['Name']
        single_channel_parms = {name: dict() for name in channel_names}
        for field_name, field_value, in info_dict.items():
            for channel_name, value in zip(channel_names, field_value):
                single_channel_parms[channel_name][field_name] = value
        for value in single_channel_parms.values():
            if value['Direction'] == 'both':
                value['Direction'] = ['forward', 'backward']
            else:
                direction = [value['Direction']]
        scan_dir = meas_parms['scan_dir']
        for name, parms in single_channel_parms.items():
            for direction in parms['Direction']:
                key = ' '.join((name, direction))
                channel_parms[key] = dict(parms)
                channel_parms[key]['Direction'] = direction
                data = signal_dict[name][direction]
                if scan_dir == 'up':
                    data = np.flip(data, axis=0)
                if direction == 'backward':
                    data = np.flip(data, axis=1)
                data_dict[key] = data
        parm_dict['channel_parms'] = channel_parms

        # Position dimensions
        num_cols, num_rows = header_dict['scan_pixels']
        width, height = header_dict['scan_range']
        pos_names = ['X', 'Y']
        pos_units = ['nm', 'nm']
        pos_vals = np.vstack([
            np.linspace(0, width, num_cols),
            np.linspace(0, height, num_rows),
        ])
        pos_vals *= 1e9
        dims = [Dimension(values, name=name, quantity='Length', units=unit,
                          dimension_type=DimensionType.SPATIAL) for
                name, unit, values
                in zip(pos_names, pos_units, pos_vals)]
        data_dict['Dimensions'] = dims

        return parm_dict, data_dict

  
    def read(self):
        """
        Reads data from .sxm files into sidpy.Dataset objects
        Note that multiple channels are treated as separate dataset objects,
        Thus returning a list of length N where N is the number of channels.

        Returns
        -------
        dataset_list: (list) of sidpy.Dataset objects
        """
       
        reader = nap.read.Scan
       
        nanonis_data = reader(self._input_file_path)

        header_dict = nanonis_data.header
        signal_dict = nanonis_data.signals

        parm_dict, data_dict = self._parse_sxm_parms(header_dict,
                                                         signal_dict)
        
        self.parm_dict = parm_dict
        self.data_dict = data_dict

        #Specify dimensions
        x_dim = self.data_dict['Dimensions'][0]
        y_dim = self.data_dict['Dimensions'][1]

        dataset_list = []
        channel_parms = self.parm_dict['channel_parms']

        for dataset_name in list(self.data_dict.keys())[:-1]:
            
            data_mat = self.data_dict[dataset_name]
            
            #Make a sidpy dataset
            data_set = sid.Dataset.from_array(data_mat, title = dataset_name)

            #Set the data type
            data_set.data_type = sid.DataType.IMAGE
            
            metadata = channel_parms[dataset_name]

            # Add quantity and units
            data_set.units = metadata['Unit']
            data_set.quantity = metadata['Name']

            # Add dimension info
            data_set.set_dimension(0, x_dim)
            data_set.set_dimension(1, y_dim)
        
            # append metadata 
            def merge_dict(dict1, dict2):
                res = {**dict1, **dict2}
                return res
            
            chan_metadata = self.parm_dict['channel_parms'][dataset_name]
            orig_metadata = self.parm_dict['meas_parms']
            
            data_set.metadata =  merge_dict(chan_metadata,orig_metadata)
            dataset_list.append(data_set)
        
        return dataset_list

    def can_read(self):
        """
        Tests whether or not the provided file has a .dm3 extension
        Returns
        -------
        """
        # TODO: Add dat eventually
        if nap is None:
            return False
        return super(NanonisSXMReader, self).can_read(extension='sxm')


#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-

################################################################################
# Python class for reading Nion Swift files into sidpy Dataset
# and extracting all metadata
#
# Written by Gerd Duscher, UTK 2020
#
# Works for python 3
#
################################################################################
from __future__ import division, print_function, absolute_import, unicode_literals

import json
import struct
import h5py
# from warnings import warn
import sys
import numpy as np
import os

import sidpy

__all__ = ["NionReader", "version"]

version = '0.1beta'

debugLevel = 0  # 0=none, 1-3=basic, 4-5=simple, 6-10 verbose

if sys.version_info.major == 3:
    unicode = str

# ### utility functions ###


def parse_zip(fp):
    """
        Parse the zip file headers at fp
        :param fp: the file pointer from which to parse the zip file
        :return: A tuple of local files, directory headers, and end of central directory
        The local files are dictionary where the keys are the local file offset and the
        values are each a tuple consisting of the name, data position, data length, and crc32.
        The directory headers are a dictionary where the keys are the names of the files
        and the values are a tuple consisting of the directory header position, and the
        associated local file position.
        The end of central directory is a tuple consisting of the location of the end of
        central directory header and the location of the first directory header.
        This method will seek to location 0 of fp and leave fp at end of file.

        This function is copied from  nionswift/nion/swift/model/NDataHandler.py

    """
    local_files = {}
    dir_files = {}
    eocd = None
    fp.seek(0)
    while True:
        pos = fp.tell()
        signature = struct.unpack('I', fp.read(4))[0]
        if signature == 0x04034b50:
            fp.seek(pos + 14)
            crc32 = struct.unpack('I', fp.read(4))[0]
            fp.seek(pos + 18)
            data_len = struct.unpack('I', fp.read(4))[0]
            fp.seek(pos + 26)
            name_len = struct.unpack('H', fp.read(2))[0]
            extra_len = struct.unpack('H', fp.read(2))[0]
            name_bytes = fp.read(name_len)
            fp.seek(extra_len, os.SEEK_CUR)
            data_pos = fp.tell()
            fp.seek(data_len, os.SEEK_CUR)
            local_files[pos] = (name_bytes, data_pos, data_len, crc32)
        elif signature == 0x02014b50:
            fp.seek(pos + 28)
            name_len = struct.unpack('H', fp.read(2))[0]
            extra_len = struct.unpack('H', fp.read(2))[0]
            comment_len = struct.unpack('H', fp.read(2))[0]
            fp.seek(pos + 42)
            pos2 = struct.unpack('I', fp.read(4))[0]
            name_bytes = fp.read(name_len)
            fp.seek(pos + 46 + name_len + extra_len + comment_len)
            dir_files[name_bytes] = (pos, pos2)
        elif signature == 0x06054b50:
            fp.seek(pos + 16)
            pos2 = struct.unpack('I', fp.read(4))[0]
            eocd = (pos, pos2)
            break
        else:
            raise IOError()
    return local_files, dir_files, eocd


class NionReader(sidpy.Reader):

    def __init__(self, file_path, verbose=False):
        """
        file_path: filepath to dm3 file.
        """

        super().__init__(file_path)

        # initialize variables ##
        self.verbose = verbose
        self.__filename = file_path

        path, file_name = os.path.split(self.__filename)
        self.basename, self.extension = os.path.splitext(file_name)
        self.data_cube = None
        self.original_metadata = {}
        self.dimensions = []
        if 'ndata' in self.extension:

            # - open file for reading
            try:
                self.__f = open(self.__filename, "rb")
            except FileNotFoundError:
                raise FileNotFoundError('File not found')
            try:
                local_files, dir_files, eocd = parse_zip(self.__f)
            except IOError:
                raise IOError("File {} does not seem to be of Nion`s .ndata format".format(self.__filename))
            self.__f.close()
        elif self.extension == '.h5':
            try:
                fp = h5py.File(self.__filename, mode='a')
                if 'data' not in fp:
                    raise IOError("File {} does not seem to be of Nion`s .h5 format".format(self.__filename))
                fp.close()
            except IOError:
                raise IOError("File {} does not seem to be of Nion`s .h5 format".format(self.__filename))

    def read(self):
        if 'ndata' in self.extension:
            try:
                self.__f = open(self.__filename, "rb")
            except FileNotFoundError:
                raise FileNotFoundError('File not found')
            local_files, dir_files, eocd = parse_zip(self.__f)

            contains_data = b"data.npy" in dir_files
            contains_metadata = b"metadata.json" in dir_files
            file_count = contains_data + contains_metadata  # use fact that True is 1, False is 0

            self.__f.seek(local_files[dir_files[b"data.npy"][1]][1])

            self.data_cube = np.load(self.__f)

            json_pos = local_files[dir_files[b"metadata.json"][1]][1]
            json_len = local_files[dir_files[b"metadata.json"][1]][2]
            self.__f.seek(json_pos)
            json_properties = self.__f.read(json_len)

            self.original_metadata = json.loads(json_properties.decode("utf-8"))
            self.__f.close()
        elif self.extension == '.h5':
            # TODO: use lazy load for large datasets
            self.__f = h5py.File(self.__filename, 'a')
            if 'data' in self.__f:
                json_properties = self.__f['data'].attrs.get("properties", "")
                self.data_cube = self.__f['data'][:]
                self.original_metadata = json.loads(json_properties)
            self.__f.close()
        self.get_dimensions()
        # Need to switch image dimensions in Nion format
        image_dims = []
        spectral_dims = []
        for dim, axis in enumerate(self.dimensions):
            # print(dim, axis)
            if axis.dimension_type == sidpy.DimensionType.SPATIAL:
                image_dims.append(dim)
            if axis.dimension_type == sidpy.DimensionType.SPECTRAL:
                spectral_dims.append(dim)

        # convert line-scan nxN to spectral_image 1xnxN
        if len(image_dims) == 1:
            if self.data_cube.ndim > 1:
                self.data_cube = self.data_cube.reshape(1, self.data_cube.shape[0], self.data_cube.shape[1])
                new_dims = [sidpy.Dimension([1], name='x', units='pixels',
                                            quantity='distance', dimension_type='spatial'),
                            sidpy.Dimension(np.arange(self.data_cube.shape[0]), name='y', units='pixels',
                                            quantity='distance', dimension_type='spatial'),
                            self.dimensions[spectral_dims[0]]]
                self.dimensions = new_dims

        if len(image_dims) == 2:
            self.data_cube = np.swapaxes(self.data_cube, image_dims[0], image_dims[1])
            temp = self.dimensions[image_dims[0]].copy()
            self.dimensions[image_dims[0]] = self.dimensions[image_dims[1]].copy()
            self.dimensions[image_dims[1]] = temp

        dataset = sidpy.Dataset.from_array(self.data_cube)

        dim_names = []
        for dim, axis in enumerate(self.dimensions):
            dataset.set_dimension(dim, axis)

        dataset.original_metadata = self.original_metadata
        if 'dimensional_calibrations' in dataset.original_metadata:
            for dim in dataset.original_metadata['dimensional_calibrations']:
                if dim['units'] == '':
                    dim['units'] = 'pixels'

        dataset.quantity = 'intensity'
        dataset.units = 'counts'
        if 'description' in dataset.original_metadata:
            dataset.title = dataset.original_metadata['description']['title']
        else:
            if 'title' in dataset.original_metadata:
                dataset.title = dataset.original_metadata['title']
            else:
                path, file_name = os.path.split(self.__filename)
                basename, extension = os.path.splitext(file_name)
                dataset.title = basename

        if 'data_source' in dataset.original_metadata:
            dataset.source = dataset.original_metadata['data_source']
        else:
            dataset.source = 'NionReader'

        self.set_data_type(dataset)
        dataset.modality = 'STEM data'
        dataset.h5_dataset = None

        return dataset

    def set_data_type(self, dataset):

        spectral_dim = False
        for axis in dataset._axes.values():
            if axis.dimension_type == sidpy.DimensionType.SPECTRAL:
                spectral_dim = True

        if len(dataset.shape) > 3:
            raise NotImplementedError('Data_type not implemented yet')
        elif len(dataset.shape) == 3:
            if spectral_dim:
                dataset.data_type = 'spectral_image'
            else:
                dataset.data_type = 'IMAGE_STACK'
                for dim, axis in dataset._axes.items():
                    if axis.dimension_type != sidpy.DimensionType.SPATIAL:
                        dataset.set_dimension(dim, sidpy.Dimension(axis.values,
                                                                   name='frame',
                                                                   units='frame',
                                                                   quantity='stack',
                                                                   dimension_type=sidpy.DimensionType.TEMPORAL))
                        break

        elif len(dataset.shape) == 2:
            if spectral_dim:
                dataset.data_type = sidpy.DataType.SPECTRAL_IMAGE
            else:
                dataset.data_type = sidpy.DataType.IMAGE
        elif len(dataset.shape) == 1:
            if spectral_dim:
                dataset.data_type = sidpy.DataType.SPECTRUM
            else:
                dataset.data_type = sidpy.DataType.LINE_PLOT

    def get_dimensions(self):
        dic = self.original_metadata

        reciprocal_name = 'u'
        spatial_name = 'x'

        if 'dimensional_calibrations' in dic:
            dimension_list = dic['dimensional_calibrations']
        elif 'spatial_calibrations' in dic:
            dimension_list = dic['spatial_calibrations']
        else:
            return

        for dim in range(len(dimension_list)):
            dimension_tags = dimension_list[dim]
            units = dimension_tags['units']
            values = (np.arange(self.data_cube.shape[int(dim)])-dimension_tags['offset']) * dimension_tags['scale']

            if 'eV' == units:
                self.dimensions.append(sidpy.Dimension(values, name='energy_loss', units=units,
                                                       quantity='energy-loss', dimension_type='spectral'))
            elif 'eV' in units:
                self.dimensions.append(sidpy.Dimension(values, name='energy', units=units,
                                                       quantity='energy', dimension_type='spectral'))
            elif '1/' in units or units in ['mrad', 'rad']:
                self.dimensions.append(sidpy.Dimension(values, name=reciprocal_name, units=units,
                                                       quantity='reciprocal distance',
                                                       dimension_type='reciprocal'))
                reciprocal_name = chr(ord(reciprocal_name) + 1)
            elif 'nm' in units:
                self.dimensions.append(sidpy.Dimension(values, name=spatial_name, units=units,
                                                       quantity='distance', dimension_type='spatial'))
                spatial_name = chr(ord(spatial_name) + 1)
            else:
                self.dimensions.append(sidpy.Dimension(values, name='generic_{}'.format(dim), units='generic',
                                                       quantity='generic', dimension_type='UNKNOWN'))

    def get_filename(self):
        return self.__filename

    filename = property(get_filename)

    def get_raw(self):
        return self.data

    data = property(get_raw)

    def get_tags(self):
        return self.original_metadata

    tags = property(get_tags)



from codecs import open
import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.rst')) as f:
    long_description = f.read()

with open(os.path.join(here, 'SciFiReaders/__version__.py')) as f:
    __version__ = f.read().split("'")[1]

# TODO: Move requirements to requirements.txt
requirements = [  # basic
                'numpy>=1.10',
                'toolz',  # dask installation failing without this
                'cytoolz',  # dask installation failing without this
                'dask>=2.20.0',
                'sidpy>=0.0.2',
                # generic:
                'pillow>=5.4.0',
                # Reader specific ones go to extras
               ]

setup(
    name='SciFiReaders',
    version=__version__,
    description='Tools for extracting data and metadata from scientific data '
                'files',
    long_description=long_description,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Scientific/Engineering :: Information Analysis'],
    keywords=['imaging', 'spectra', 'multidimensional', 'scientific'],
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*",
                                    "tests"]),
    url='https://pycroscopy.github.io/SciFiReaders/about.html',
    license='MIT',
    author='Pycroscopy contributors',
    author_email='pycroscopy@gmail.com',
    install_requires=requirements,
    setup_requires=['pytest-runner'],
    tests_require=['pytest', 'pywget', 'hyperspy'],
    platforms=['Linux', 'Mac OSX', 'Windows 10/8.1/8/7'],
    # package_data={'sample':['dataset_1.dat']}
    test_suite='pytest',
    # dependency='',
    # dependency_links=[''],
    include_package_data=True,
    # https://setuptools.readthedocs.io/en/latest/setuptools.html#declaring-dependencies
    extras_require={
        'hyperspy':  ["hyperspy"],
        'igor': ["igor"],
        "gwyddion": ["gwyfile"],
        'nanonispy': ['nanonispy'],
        'sid': ['pyUSID', 'pyNSID']
    },
    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    # package_data={
    #     'sample': ['package_data.dat'],
    # },

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[('my_data', ['data/data_file'])],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    # entry_points={
    #     'console_scripts': [
    #         'sample=sample:main',
    #     ],
    # },
)
